@@ -4,18 +4,21 @@ Name Server and helper functions.
 Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
 """
 
-from __future__ import with_statement
+from __future__ import print_function
 import warnings
 import re
 import logging
 import socket
 import sys
-from Pyro4.threadutil import RLock, Thread
+import time
+import threading
 from Pyro4.errors import NamingError, PyroError, ProtocolError
-from Pyro4 import core, socketutil
-import Pyro4.constants
+from Pyro4 import core, socketutil, constants
+from Pyro4.configuration import config
+from Pyro4.core import _locateNS as locateNS, _resolve as resolve    # API compatibility with older versions
 
-__all__ = ["locateNS", "resolve", "type_meta", "startNS", "startNSloop", "MemoryStorage"]
+
+__all__ = ["locateNS", "resolve", "type_meta", "startNSloop", "startNS"]
 
 if sys.version_info >= (3, 0):
     basestring = str
@@ -61,6 +64,7 @@ class MemoryStorage(dict):
         pass
 
 
+@core.expose
 class NameServer(object):
     """
     Pyro name server. Provides a simple flat name space to map logical object names to Pyro URIs.
@@ -71,7 +75,7 @@ class NameServer(object):
         if storageProvider is None:
             self.storage = MemoryStorage()
             log.debug("using volatile in-memory dict storage")
-        self.lock = RLock()
+        self.lock = threading.RLock()
 
     def count(self):
         return len(self.storage)
@@ -84,7 +88,10 @@ class NameServer(object):
         try:
             uri, metadata = self.storage[name]
             uri = core.URI(uri)
-            return (uri, metadata or frozenset()) if return_metadata else uri
+            if return_metadata:
+                metadata = list(metadata) if metadata else []
+                return uri, metadata
+            return uri
         except KeyError:
             raise NamingError("unknown name: " + name)
 
@@ -127,20 +134,20 @@ class NameServer(object):
 
     def remove(self, name=None, prefix=None, regex=None):
         """Remove a registration. returns the number of items removed."""
-        if name and name in self.storage and name != Pyro4.constants.NAMESERVER_NAME:
+        if name and name in self.storage and name != constants.NAMESERVER_NAME:
             with self.lock:
                 del self.storage[name]
             return 1
         if prefix:
             items = list(self.list(prefix=prefix).keys())
-            if Pyro4.constants.NAMESERVER_NAME in items:
-                items.remove(Pyro4.constants.NAMESERVER_NAME)
+            if constants.NAMESERVER_NAME in items:
+                items.remove(constants.NAMESERVER_NAME)
             self.storage.remove_items(items)
             return len(items)
         if regex:
             items = list(self.list(regex=regex).keys())
-            if Pyro4.constants.NAMESERVER_NAME in items:
-                items.remove(Pyro4.constants.NAMESERVER_NAME)
+            if constants.NAMESERVER_NAME in items:
+                items.remove(constants.NAMESERVER_NAME)
             self.storage.remove_items(items)
             return len(items)
         return 0
@@ -150,33 +157,42 @@ class NameServer(object):
         """Retrieve the registered items as a dictionary name-to-URI. The URIs
         in the resulting dict are strings, not URI objects.
         You can filter by prefix or by regex or by metadata subset (separately)"""
+        def fix_set(result):
+            # for python 2 compatibility we cannot send sets to the default (serpent) serializer.
+            # that's why we will convert them to lists here.
+            if return_metadata:
+                fixed = {}
+                for name, data in result.items():
+                    fixed[name] = (data[0], list(data[1]))
+                return fixed
+            return result
+
         if sum(1 for x in [prefix, regex, metadata_all, metadata_any] if x is not None) > 1:
             raise ValueError("you can only filter on one thing at a time")
         with self.lock:
             if prefix:
                 result = self.storage.optimized_prefix_list(prefix, return_metadata)
                 if result is not None:
-                    return result
+                    return fix_set(result)
                 result = {}
                 for name in self.storage:
                     if name.startswith(prefix):
                         result[name] = self.storage[name] if return_metadata else self.storage[name][0]
-                return result
+                return fix_set(result)
             elif regex:
                 result = self.storage.optimized_regex_list(regex, return_metadata)
                 if result is not None:
-                    return result
+                    return fix_set(result)
                 result = {}
                 try:
                     regex = re.compile(regex)
-                except re.error:
-                    x = sys.exc_info()[1]
+                except re.error as x:
                     raise NamingError("invalid regex: " + str(x))
                 else:
                     for name in self.storage:
                         if regex.match(name):
                             result[name] = self.storage[name] if return_metadata else self.storage[name][0]
-                    return result
+                    return fix_set(result)
             elif metadata_all:
                 # return the entries which have all of the given metadata as (a subset of) their metadata
                 if isinstance(metadata_all, basestring):
@@ -184,13 +200,13 @@ class NameServer(object):
                 metadata_all and iter(metadata_all)   # validate that metadata is iterable
                 result = self.storage.optimized_metadata_search(metadata_all=metadata_all, return_metadata=return_metadata)
                 if result is not None:
-                    return result
+                    return fix_set(result)
                 metadata_all = frozenset(metadata_all)
                 result = {}
                 for name, (uri, meta) in self.storage.everything(return_metadata=True).items():
                     if metadata_all.issubset(meta):
                         result[name] = (uri, meta) if return_metadata else uri
-                return result
+                return fix_set(result)
             elif metadata_any:
                 # return the entries which have any of the given metadata as part of their metadata
                 if isinstance(metadata_any, basestring):
@@ -198,16 +214,16 @@ class NameServer(object):
                 metadata_any and iter(metadata_any)   # validate that metadata is iterable
                 result = self.storage.optimized_metadata_search(metadata_any=metadata_any, return_metadata=return_metadata)
                 if result is not None:
-                    return result
+                    return fix_set(result)
                 metadata_any = frozenset(metadata_any)
                 result = {}
                 for name, (uri, meta) in self.storage.everything(return_metadata=True).items():
                     if metadata_any & meta:
                         result[name] = (uri, meta) if return_metadata else uri
-                return result
+                return fix_set(result)
             else:
                 # just return (a copy of) everything
-                return self.storage.everything(return_metadata)
+                return fix_set(self.storage.everything(return_metadata))
 
     def ping(self):
         """A simple test method to check if the name server is running correctly."""
@@ -219,13 +235,13 @@ class NameServerDaemon(core.Daemon):
 
     def __init__(self, host=None, port=None, unixsocket=None, nathost=None, natport=None, storage=None):
         if host is None:
-            host = Pyro4.config.HOST
+            host = config.HOST
         if port is None:
-            port = Pyro4.config.NS_PORT
+            port = config.NS_PORT
         if nathost is None:
-            nathost = Pyro4.config.NATHOST
+            nathost = config.NATHOST
         if natport is None:
-            natport = Pyro4.config.NATPORT or None
+            natport = config.NATPORT or None
         storage = storage or "memory"
         if storage == "memory":
             log.debug("using volatile in-memory dict storage")
@@ -249,9 +265,18 @@ class NameServerDaemon(core.Daemon):
         if existing_count > 0:
             log.debug("number of existing entries in storage: %d", existing_count)
         super(NameServerDaemon, self).__init__(host, port, unixsocket, nathost=nathost, natport=natport)
-        self.register(self.nameserver, Pyro4.constants.NAMESERVER_NAME)
+        self.register(self.nameserver, constants.NAMESERVER_NAME)
         metadata = {"class:Pyro4.naming.NameServer"}
-        self.nameserver.register(Pyro4.constants.NAMESERVER_NAME, self.uriFor(self.nameserver), metadata=metadata)
+        self.nameserver.register(constants.NAMESERVER_NAME, self.uriFor(self.nameserver), metadata=metadata)
+        if config.NS_AUTOCLEAN > 0:
+            if not AutoCleaner.override_autoclean_min and config.NS_AUTOCLEAN < AutoCleaner.min_autoclean_value:
+                raise ValueError("NS_AUTOCLEAN cannot be smaller than " + str(AutoCleaner.min_autoclean_value))
+            log.debug("autoclean enabled")
+            self.cleaner_thread = AutoCleaner(self.nameserver)
+            self.cleaner_thread.start()
+        else:
+            log.debug("autoclean not enabled")
+            self.cleaner_thread = None
         log.info("nameserver daemon created")
 
     def close(self):
@@ -259,6 +284,10 @@ class NameServerDaemon(core.Daemon):
         if self.nameserver is not None:
             self.nameserver.storage.close()
             self.nameserver = None
+        if self.cleaner_thread:
+            self.cleaner_thread.stop = True
+            self.cleaner_thread.join()
+            self.cleaner_thread = None
 
     def __enter__(self):
         if not self.nameserver:
@@ -269,6 +298,10 @@ class NameServerDaemon(core.Daemon):
         if self.nameserver is not None:
             self.nameserver.storage.close()
         self.nameserver = None
+        if self.cleaner_thread:
+            self.cleaner_thread.stop = True
+            self.cleaner_thread.join()
+            self.cleaner_thread = None
         return super(NameServerDaemon, self).__exit__(exc_type, exc_value, traceback)
 
     def handleRequest(self, conn):
@@ -282,20 +315,79 @@ class NameServerDaemon(core.Daemon):
             raise
 
 
-class BroadcastServer(object):
-    REQUEST_NSURI = "GET_NSURI" if sys.platform == "cli" else b"GET_NSURI"
+class AutoCleaner(threading.Thread):
+    """
+    Takes care of checking every registration in the name server.
+    If it cannot be contacted anymore, it will be removed after ~20 seconds.
+    """
+    min_autoclean_value = 3
+    max_unreachable_time = 20.0
+    loop_delay = 2.0
+    override_autoclean_min = False   # only for unit test purposes
 
-    def __init__(self, nsUri, bchost=None, bcport=None):
+    def __init__(self, nameserver):
+        assert config.NS_AUTOCLEAN > 0
+        if not self.override_autoclean_min and config.NS_AUTOCLEAN < self.min_autoclean_value:
+            raise ValueError("NS_AUTOCLEAN cannot be smaller than "+str(self.min_autoclean_value))
+        super(AutoCleaner, self).__init__()
+        self.nameserver = nameserver
+        self.stop = False
+        self.daemon = True
+        self.last_cleaned = time.time()
+        self.unreachable = {}   # name->since when
+
+    def run(self):
+        while not self.stop:
+            time.sleep(self.loop_delay)
+            time_since_last_autoclean = time.time() - self.last_cleaned
+            if time_since_last_autoclean < config.NS_AUTOCLEAN:
+                continue
+            for name, uri in self.nameserver.list().items():
+                if name in (constants.DAEMON_NAME, constants.NAMESERVER_NAME):
+                    continue
+                try:
+                    uri_obj = core.URI(uri)
+                    timeout = config.COMMTIMEOUT or 5
+                    sock = socketutil.createSocket(connect=(uri_obj.host, uri_obj.port), timeout=timeout)
+                    sock.close()
+                    # if we get here, the listed server is still answering on its port
+                    if name in self.unreachable:
+                        del self.unreachable[name]
+                except socket.error:
+                    if name not in self.unreachable:
+                        self.unreachable[name] = time.time()
+                    if time.time() - self.unreachable[name] >= self.max_unreachable_time:
+                        log.info("autoclean: unregistering %s; cannot connect uri %s for %d sec", name, uri, self.max_unreachable_time)
+                        self.nameserver.remove(name)
+                        del self.unreachable[name]
+                        continue
+            self.last_cleaned = time.time()
+            if self.unreachable:
+                log.debug("autoclean: %d/%d names currently unreachable", len(self.unreachable), self.nameserver.count())
+
+
+class BroadcastServer(object):
+    class TransportServerAdapter(object):
+        # this adapter is used to be able to pass the BroadcastServer to Daemon.combine() to integrate the event loops.
+        def __init__(self, bcserver):
+            self.sockets = [bcserver]
+
+        def events(self, eventobjects):
+            for bc in eventobjects:
+                bc.processRequest()
+
+    def __init__(self, nsUri, bchost=None, bcport=None, ipv6=False):
+        self.transportServer = self.TransportServerAdapter(self)
         self.nsUri = nsUri
         if bcport is None:
-            bcport = Pyro4.config.NS_BCPORT
+            bcport = config.NS_BCPORT
         if bchost is None:
-            bchost = Pyro4.config.NS_BCHOST
-        if ":" in nsUri.host:  # ipv6
+            bchost = config.NS_BCHOST
+        if ":" in nsUri.host or ipv6:   # match nameserver's ip version
             bchost = bchost or "::"
-            self.sock = Pyro4.socketutil.createBroadcastSocket((bchost, bcport, 0, 0), reuseaddr=Pyro4.config.SOCK_REUSE, timeout=2.0)
+            self.sock = socketutil.createBroadcastSocket((bchost, bcport, 0, 0), reuseaddr=config.SOCK_REUSE, timeout=2.0)
         else:
-            self.sock = Pyro4.socketutil.createBroadcastSocket((bchost, bcport), reuseaddr=Pyro4.config.SOCK_REUSE, timeout=2.0)
+            self.sock = socketutil.createBroadcastSocket((bchost, bcport), reuseaddr=config.SOCK_REUSE, timeout=2.0)
         self._sockaddr = self.sock.getsockname()
         bchost = bchost or self._sockaddr[0]
         bcport = bcport or self._sockaddr[1]
@@ -303,7 +395,7 @@ class BroadcastServer(object):
             self.locationStr = "[%s]:%d" % (bchost, bcport)
         else:
             self.locationStr = "%s:%d" % (bchost, bcport)
-        log.info("ns broadcast server created on %s", self.locationStr)
+        log.info("ns broadcast server created on %s - %s", self.locationStr, socketutil.family_str(self.sock))
         self.running = True
 
     def close(self):
@@ -323,10 +415,11 @@ class BroadcastServer(object):
 
     def runInThread(self):
         """Run the broadcast server loop in its own thread."""
-        thread = Thread(target=self.__requestLoop)
+        thread = threading.Thread(target=self.__requestLoop)
         thread.setDaemon(True)
         thread.start()
         log.debug("broadcast server loop running in own thread")
+        return thread
 
     def __requestLoop(self):
         while self.running:
@@ -336,7 +429,7 @@ class BroadcastServer(object):
     def processRequest(self):
         try:
             data, addr = self.sock.recvfrom(100)
-            if data == self.REQUEST_NSURI:
+            if data == b"GET_NSURI":
                 responsedata = core.URI(self.nsUri)
                 if responsedata.host == "0.0.0.0":
                     # replace INADDR_ANY address by the interface IP address that connects to the requesting client
@@ -351,7 +444,7 @@ class BroadcastServer(object):
         except socket.error:
             pass
         except SystemError:
-            if sys.platform == 'cli' and not self.running:
+            if sys.platform == "cli" and not self.running:
                 # ironpython throws these systemerrors when shutting down... we can ignore them.
                 pass
             else:
@@ -375,7 +468,11 @@ def startNSloop(host=None, port=None, enableBroadcast=True, bchost=None, bcport=
         hostip = "Unix domain socket"
     else:
         hostip = daemon.sock.getsockname()[0]
-        if hostip.startswith("127."):
+        if daemon.sock.family == socket.AF_INET6:       # ipv6 doesn't have broadcast. We should probably use multicast instead...
+            print("Not starting broadcast server for IPv6.")
+            log.info("Not starting NS broadcast server because NS is using IPv6")
+            enableBroadcast = False
+        elif hostip.startswith("127.") or hostip == "::1":
             print("Not starting broadcast server for localhost.")
             log.info("Not starting NS broadcast server because NS is bound to localhost")
             enableBroadcast = False
@@ -383,7 +480,7 @@ def startNSloop(host=None, port=None, enableBroadcast=True, bchost=None, bcport=
             # Make sure to pass the internal uri to the broadcast responder.
             # It is almost always useless to let it return the external uri,
             # because external systems won't be able to talk to this thing anyway.
-            bcserver = BroadcastServer(internalUri, bchost, bcport)
+            bcserver = BroadcastServer(internalUri, bchost, bcport, ipv6=daemon.sock.family==socket.AF_INET6)
             print("Broadcast server running on %s" % bcserver.locationStr)
             bcserver.runInThread()
     existing = daemon.nameserver.count()
@@ -420,115 +517,10 @@ def startNS(host=None, port=None, enableBroadcast=True, bchost=None, bcport=None
             enableBroadcast = False
         if enableBroadcast:
             internalUri = daemon.uriFor(daemon.nameserver, nat=False)
-            bcserver = BroadcastServer(internalUri, bchost, bcport)
+            bcserver = BroadcastServer(internalUri, bchost, bcport, ipv6=daemon.sock.family==socket.AF_INET6)
     return nsUri, daemon, bcserver
 
 
-def locateNS(host=None, port=None, broadcast=True, hmac_key=None):
-    """Get a proxy for a name server somewhere in the network."""
-    if host is None:
-        # first try localhost if we have a good chance of finding it there
-        if Pyro4.config.NS_HOST in ("localhost", "::1") or Pyro4.config.NS_HOST.startswith("127."):
-            host = Pyro4.config.NS_HOST
-            if ":" in host:  # ipv6
-                host = "[%s]" % host
-            uristring = "PYRO:%s@%s:%d" % (Pyro4.constants.NAMESERVER_NAME, host, port or Pyro4.config.NS_PORT)
-            log.debug("locating the NS: %s", uristring)
-            proxy = core.Proxy(uristring)
-            proxy._pyroHmacKey = hmac_key
-            try:
-                proxy.ping()
-                log.debug("located NS")
-                return proxy
-            except PyroError:
-                pass
-        if broadcast:
-            # broadcast lookup
-            if not port:
-                port = Pyro4.config.NS_BCPORT
-            log.debug("broadcast locate")
-            sock = Pyro4.socketutil.createBroadcastSocket(reuseaddr=Pyro4.config.SOCK_REUSE, timeout=0.7)
-            for _ in range(3):
-                try:
-                    for bcaddr in Pyro4.config.parseAddressesString(Pyro4.config.BROADCAST_ADDRS):
-                        try:
-                            sock.sendto(BroadcastServer.REQUEST_NSURI, 0, (bcaddr, port))
-                        except socket.error:
-                            x = sys.exc_info()[1]
-                            err = getattr(x, "errno", x.args[0])
-                            # handle some errno's that some platforms like to throw:
-                            if err not in Pyro4.socketutil.ERRNO_EADDRNOTAVAIL and err not in Pyro4.socketutil.ERRNO_EADDRINUSE:
-                                raise
-                    data, _ = sock.recvfrom(100)
-                    sock.close()
-                    if sys.version_info >= (3, 0):
-                        data = data.decode("iso-8859-1")
-                    log.debug("located NS: %s", data)
-                    proxy = core.Proxy(data)
-                    proxy._pyroHmacKey = hmac_key
-                    return proxy
-                except socket.timeout:
-                    continue
-            try:
-                sock.shutdown(socket.SHUT_RDWR)
-            except (OSError, socket.error):
-                pass
-            sock.close()
-            log.debug("broadcast locate failed, try direct connection on NS_HOST")
-        else:
-            log.debug("skipping broadcast lookup")
-        # broadcast failed or skipped, try PYRO directly on specific host
-        host = Pyro4.config.NS_HOST
-        port = Pyro4.config.NS_PORT
-    # pyro direct lookup
-    if not port:
-        port = Pyro4.config.NS_PORT
-    if ":" in host:
-        host = "[%s]" % host
-    if core.URI.isUnixsockLocation(host):
-        uristring = "PYRO:%s@%s" % (Pyro4.constants.NAMESERVER_NAME, host)
-    else:
-        uristring = "PYRO:%s@%s:%d" % (Pyro4.constants.NAMESERVER_NAME, host, port)
-    uri = core.URI(uristring)
-    log.debug("locating the NS: %s", uri)
-    proxy = core.Proxy(uri)
-    proxy._pyroHmacKey = hmac_key
-    try:
-        proxy.ping()
-        log.debug("located NS")
-        return proxy
-    except PyroError as x:
-        e = NamingError("Failed to locate the nameserver")
-        if sys.version_info >= (3, 0):
-            e.__cause__ = x
-        raise e
-
-
-def resolve(uri, hmac_key=None):
-    """
-    Resolve a 'magic' uri (PYRONAME) into the direct PYRO uri.
-    It finds a name server, and use that to resolve a PYRONAME uri into the direct PYRO uri pointing to the named object.
-    If uri is already a PYRO uri, it is returned unmodified.
-    You can consider this a shortcut function so that you don't have to locate and use a name server proxy yourself.
-    Note: if you need to resolve more than a few names, consider using the name server directly instead of repeatedly
-    calling this function, to avoid the name server lookup overhead from each call.
-    """
-    if isinstance(uri, basestring):
-        uri = core.URI(uri)
-    elif not isinstance(uri, core.URI):
-        raise TypeError("can only resolve Pyro URIs")
-    if uri.protocol == "PYRO":
-        return uri
-    log.debug("resolving %s", uri)
-    if uri.protocol == "PYRONAME":
-        nameserver = locateNS(uri.host, uri.port, hmac_key=hmac_key)
-        uri = nameserver.lookup(uri.object)
-        nameserver._pyroRelease()
-        return uri
-    else:
-        raise PyroError("invalid uri protocol")
-
-
 def type_meta(class_or_object, prefix="class:"):
     """extracts type metadata from the given class or object, can be used as Name server metadata."""
     if hasattr(class_or_object, "__mro__"):
