@@ -1,17 +1,22 @@
 """
-Miscellaneous utilities.
+Miscellaneous utilities, and serializers.
 
 Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
 """
 
 import sys
 import zlib
+import uuid
 import logging
 import linecache
 import traceback
 import inspect
-import Pyro4.errors
-import Pyro4.message
+import struct
+import datetime
+import decimal
+import numbers
+from Pyro4 import errors
+from Pyro4.configuration import config
 
 try:
     import copyreg
@@ -48,7 +53,7 @@ def getPyroTraceback(ex_type=None, ex_value=None, ex_tb=None):
             ex_type, ex_value, ex_tb = sys.exc_info()
 
         remote_tb = getattr(ex_value, "_pyroTraceback", None)
-        local_tb = formatTraceback(ex_type, ex_value, ex_tb, Pyro4.config.DETAILED_TRACEBACK)
+        local_tb = formatTraceback(ex_type, ex_value, ex_tb, config.DETAILED_TRACEBACK)
         if remote_tb:
             remote_tb = formatRemoteTraceback(remote_tb)
             return local_tb + remote_tb
@@ -139,8 +144,9 @@ else:
     for name, t in vars(builtins).items():
         if type(t) is type and issubclass(t, BaseException):
             all_exceptions[name] = t
-for name, t in vars(Pyro4.errors).items():
-    if type(t) is type and issubclass(t, Pyro4.errors.PyroError):
+    buffer = bytearray
+for name, t in vars(errors).items():
+    if type(t) is type and issubclass(t, errors.PyroError):
         all_exceptions[name] = t
 
 
@@ -158,6 +164,8 @@ class SerializerBase(object):
     def deserializeData(self, data, compressed=False):
         """Deserializes the given data (bytes). Set compressed to True to decompress the data first."""
         if compressed:
+            if sys.version_info < (3, 0):
+                data = self._convertToBytes(data)
             data = zlib.decompress(data)
         return self.loads(data)
 
@@ -171,6 +179,8 @@ class SerializerBase(object):
         """Deserializes the given call data back to (object, method, vargs, kwargs) tuple.
         Set compressed to True to decompress the data first."""
         if compressed:
+            if sys.version_info < (3, 0):
+                data = self._convertToBytes(data)
             data = zlib.decompress(data)
         return self.loadsCall(data)
 
@@ -186,6 +196,15 @@ class SerializerBase(object):
     def dumpsCall(self, obj, method, vargs, kwargs):
         raise NotImplementedError("implement in subclass")
 
+    def _convertToBytes(self, data):
+        t = type(data)
+        if t is not bytes:
+            if t in (bytearray, buffer):
+                return bytes(data)
+            if t is memoryview:
+                return data.tobytes()
+        return data
+
     def __compressdata(self, data, compress):
         if not compress or len(data) < 200:
             return data, False  # don't waste time compressing small messages
@@ -194,13 +213,6 @@ class SerializerBase(object):
             return compressed, True
         return data, False
 
-    @classmethod
-    def decompress_if_needed(cls, message):
-        if message.flags & Pyro4.message.FLAGS_COMPRESSED:
-            message.data = zlib.decompress(message.data)
-            message.flags &= ~Pyro4.message.FLAGS_COMPRESSED
-        return message
-
     @classmethod
     def register_type_replacement(cls, object_type, replacement_function):
         raise NotImplementedError("implement in subclass")
@@ -220,7 +232,7 @@ class SerializerBase(object):
                     serializer.ser_builtins_dict(d, stream, level)
 
                 serpent.register_class(clazz, serpent_converter)
-            except Pyro4.errors.ProtocolError:
+            except errors.ProtocolError:
                 pass
 
     @classmethod
@@ -233,7 +245,7 @@ class SerializerBase(object):
             get_serializer_by_id(SerpentSerializer.serializer_id)
             import serpent
             serpent.unregister_class(clazz)
-        except Pyro4.errors.ProtocolError:
+        except errors.ProtocolError:
             pass
 
     @classmethod
@@ -299,7 +311,7 @@ class SerializerBase(object):
                 value["__class__"] = obj.__class__.__module__ + "." + obj.__class__.__name__
                 return value
             else:
-                raise Pyro4.errors.SerializeError("don't know how to serialize class " + str(obj.__class__) + " using serializer " + str(cls.__name__) + ". Give it vars() or an appropriate __getstate__")
+                raise errors.SerializeError("don't know how to serialize class " + str(obj.__class__) + " using serializer " + str(cls.__name__) + ". Give it vars() or an appropriate __getstate__")
 
     @classmethod
     def dict_to_class(cls, data):
@@ -308,6 +320,7 @@ class SerializerBase(object):
         Only a fixed set of classes are recognized.
         Not used for the pickle serializer.
         """
+        from Pyro4 import core, futures  # XXX circular
         classname = data.get("__class__", "<unknown>")
         if isinstance(classname, bytes):
             classname = classname.decode("utf-8")
@@ -315,37 +328,43 @@ class SerializerBase(object):
             converter = cls.__custom_dict_to_class_registry[classname]
             return converter(classname, data)
         if "__" in classname:
-            raise Pyro4.errors.SecurityError("refused to deserialize types with double underscores in their name: " + classname)
-        # because of efficiency reasons the constructors below are hardcoded here instead of added on a per-class basis to the dict-to-class registry
+            raise errors.SecurityError("refused to deserialize types with double underscores in their name: " + classname)
+        # for performance, the constructors below are hardcoded here instead of added on a per-class basis to the dict-to-class registry
         if classname.startswith("Pyro4.core."):
             if classname == "Pyro4.core.URI":
-                uri = Pyro4.core.URI.__new__(Pyro4.core.URI)
+                uri = core.URI.__new__(core.URI)
                 uri.__setstate_from_dict__(data["state"])
                 return uri
             elif classname == "Pyro4.core.Proxy":
-                proxy = Pyro4.core.Proxy.__new__(Pyro4.core.Proxy)
+                proxy = core.Proxy.__new__(core.Proxy)
                 proxy.__setstate_from_dict__(data["state"])
                 return proxy
             elif classname == "Pyro4.core.Daemon":
-                daemon = Pyro4.core.Daemon.__new__(Pyro4.core.Daemon)
+                daemon = core.Daemon.__new__(core.Daemon)
                 daemon.__setstate_from_dict__(data["state"])
                 return daemon
         elif classname.startswith("Pyro4.util."):
-            if classname == "Pyro4.util.PickleSerializer":
+            if classname == "Pyro4.util.SerpentSerializer":
+                return SerpentSerializer()
+            elif classname == "Pyro4.util.PickleSerializer":
                 return PickleSerializer()
             elif classname == "Pyro4.util.MarshalSerializer":
                 return MarshalSerializer()
             elif classname == "Pyro4.util.JsonSerializer":
                 return JsonSerializer()
-            elif classname == "Pyro4.util.SerpentSerializer":
-                return SerpentSerializer()
+            elif classname == "Pyro4.util.MsgpackSerializer":
+                return MsgpackSerializer()
+            elif classname == "Pyro4.util.DillSerializer":
+                return DillSerializer()
         elif classname.startswith("Pyro4.errors."):
-            errortype = getattr(Pyro4.errors, classname.split('.', 2)[2])
-            if issubclass(errortype, Pyro4.errors.PyroError):
+            errortype = getattr(errors, classname.split('.', 2)[2])
+            if issubclass(errortype, errors.PyroError):
                 return SerializerBase.make_exception(errortype, data)
         elif classname == "Pyro4.futures._ExceptionWrapper":
-            ex = SerializerBase.dict_to_class(data["exception"])
-            return Pyro4.futures._ExceptionWrapper(ex)
+            ex = data["exception"]
+            if isinstance(ex, dict) and "__class__" in ex:
+                ex = SerializerBase.dict_to_class(ex)
+            return futures._ExceptionWrapper(ex)
         elif data.get("__exception__", False):
             if classname in all_exceptions:
                 return SerializerBase.make_exception(all_exceptions[classname], data)
@@ -367,13 +386,8 @@ class SerializerBase(object):
                 exceptiontype = getattr(sqlite3, short_classname)
                 if issubclass(exceptiontype, BaseException):
                     return SerializerBase.make_exception(exceptiontype, data)
-
-        # try one of the serializer classes
-        for serializer in _serializers.values():
-            if classname == serializer.__class__.__name__:
-                return serializer
         log.warning("unsupported serialized class: " + classname)
-        raise Pyro4.errors.SerializeError("unsupported serialized class: " + classname)
+        raise errors.SerializeError("unsupported serialized class: " + classname)
 
     @staticmethod
     def make_exception(exceptiontype, data):
@@ -387,7 +401,7 @@ class SerializerBase(object):
     def recreate_classes(self, literal):
         t = type(literal)
         if t is set:
-            return set([self.recreate_classes(x) for x in literal])
+            return {self.recreate_classes(x) for x in literal}
         if t is list:
             return [self.recreate_classes(x) for x in literal]
         if t is tuple:
@@ -416,18 +430,20 @@ class PickleSerializer(SerializerBase):
     A (de)serializer that wraps the Pickle serialization protocol.
     It can optionally compress the serialized data, and is thread safe.
     """
-    serializer_id = Pyro4.message.SERIALIZER_PICKLE
+    serializer_id = 4  # never change this
 
     def dumpsCall(self, obj, method, vargs, kwargs):
-        return pickle.dumps((obj, method, vargs, kwargs), Pyro4.config.PICKLE_PROTOCOL_VERSION)
+        return pickle.dumps((obj, method, vargs, kwargs), config.PICKLE_PROTOCOL_VERSION)
 
     def dumps(self, data):
-        return pickle.dumps(data, Pyro4.config.PICKLE_PROTOCOL_VERSION)
+        return pickle.dumps(data, config.PICKLE_PROTOCOL_VERSION)
 
     def loadsCall(self, data):
+        data = self._convertToBytes(data)
         return pickle.loads(data)
 
     def loads(self, data):
+        data = self._convertToBytes(data)
         return pickle.loads(data)
 
     @classmethod
@@ -441,9 +457,39 @@ class PickleSerializer(SerializerBase):
             pass
 
 
+class DillSerializer(SerializerBase):
+    """
+    A (de)serializer that wraps the Dill serialization protocol.
+    It can optionally compress the serialized data, and is thread safe.
+    """
+    serializer_id = 5  # never change this
+
+    def dumpsCall(self, obj, method, vargs, kwargs):
+        return dill.dumps((obj, method, vargs, kwargs), config.DILL_PROTOCOL_VERSION)
+
+    def dumps(self, data):
+        return dill.dumps(data, config.DILL_PROTOCOL_VERSION)
+
+    def loadsCall(self, data):
+        return dill.loads(data)
+
+    def loads(self, data):
+        return dill.loads(data)
+
+    @classmethod
+    def register_type_replacement(cls, object_type, replacement_function):
+        def copyreg_function(obj):
+            return replacement_function(obj).__reduce__()
+
+        try:
+            copyreg.pickle(object_type, copyreg_function)
+        except TypeError:
+            pass
+
+
 class MarshalSerializer(SerializerBase):
     """(de)serializer that wraps the marshal serialization protocol."""
-    serializer_id = Pyro4.message.SERIALIZER_MARSHAL
+    serializer_id = 3  # never change this
 
     def dumpsCall(self, obj, method, vargs, kwargs):
         return marshal.dumps((obj, method, vargs, kwargs))
@@ -454,14 +500,38 @@ class MarshalSerializer(SerializerBase):
         except (ValueError, TypeError):
             return marshal.dumps(self.class_to_dict(data))
 
-    def loadsCall(self, data):
-        obj, method, vargs, kwargs = marshal.loads(data)
-        vargs = self.recreate_classes(vargs)
-        kwargs = self.recreate_classes(kwargs)
-        return obj, method, vargs, kwargs
+    if sys.platform == "cli":
+        def loadsCall(self, data):
+            if type(data) is not str:
+                # Ironpython's marshal expects str...
+                data = str(data)
+            obj, method, vargs, kwargs = marshal.loads(data)
+            vargs = self.recreate_classes(vargs)
+            kwargs = self.recreate_classes(kwargs)
+            return obj, method, vargs, kwargs
 
-    def loads(self, data):
-        return self.recreate_classes(marshal.loads(data))
+        def loads(self, data):
+            if type(data) is not str:
+                # Ironpython's marshal expects str...
+                data = str(data)
+            return self.recreate_classes(marshal.loads(data))
+    else:
+        def loadsCall(self, data):
+            data = self._convertToBytes(data)
+            obj, method, vargs, kwargs = marshal.loads(data)
+            vargs = self.recreate_classes(vargs)
+            kwargs = self.recreate_classes(kwargs)
+            return obj, method, vargs, kwargs
+
+        def loads(self, data):
+            data = self._convertToBytes(data)
+            return self.recreate_classes(marshal.loads(data))
+
+    @classmethod
+    def class_to_dict(cls, obj):
+        if isinstance(obj, uuid.UUID):
+            return str(obj)
+        return super(MarshalSerializer, cls).class_to_dict(obj)
 
     @classmethod
     def register_type_replacement(cls, object_type, replacement_function):
@@ -470,7 +540,7 @@ class MarshalSerializer(SerializerBase):
 
 class SerpentSerializer(SerializerBase):
     """(de)serializer that wraps the serpent serialization protocol."""
-    serializer_id = Pyro4.message.SERIALIZER_SERPENT
+    serializer_id = 1  # never change this
 
     def dumpsCall(self, obj, method, vargs, kwargs):
         return serpent.dumps((obj, method, vargs, kwargs), module_in_classname=True)
@@ -498,10 +568,16 @@ class SerpentSerializer(SerializerBase):
 
         serpent.register_class(object_type, custom_serializer)
 
+    @classmethod
+    def dict_to_class(cls, data):
+        if data.get("__class__") == "float":
+            return float(data["value"])     # serpent encodes a float nan as a special class dict like this
+        return super(SerpentSerializer, cls).dict_to_class(data)
+
 
 class JsonSerializer(SerializerBase):
     """(de)serializer that wraps the json serialization protocol."""
-    serializer_id = Pyro4.message.SERIALIZER_JSON
+    serializer_id = 2  # never change this
 
     __type_replacements = {}
 
@@ -515,14 +591,14 @@ class JsonSerializer(SerializerBase):
         return data.encode("utf-8")
 
     def loadsCall(self, data):
-        data = data.decode("utf-8")
+        data = self._convertToBytes(data).decode("utf-8")
         data = json.loads(data)
         vargs = self.recreate_classes(data["params"])
         kwargs = self.recreate_classes(data["kwargs"])
         return data["object"], data["method"], vargs, kwargs
 
     def loads(self, data):
-        data = data.decode("utf-8")
+        data = self._convertToBytes(data).decode("utf-8")
         return self.recreate_classes(json.loads(data))
 
     def default(self, obj):
@@ -531,6 +607,12 @@ class JsonSerializer(SerializerBase):
             obj = replacer(obj)
         if isinstance(obj, set):
             return tuple(obj)  # json module can't deal with sets so we make a tuple out of it
+        if isinstance(obj, uuid.UUID):
+            return str(obj)
+        if isinstance(obj, (datetime.datetime, datetime.date)):
+            return obj.isoformat()
+        if isinstance(obj, decimal.Decimal):
+            return str(obj)
         return self.class_to_dict(obj)
 
     @classmethod
@@ -538,6 +620,73 @@ class JsonSerializer(SerializerBase):
         cls.__type_replacements[object_type] = replacement_function
 
 
+class MsgpackSerializer(SerializerBase):
+    """(de)serializer that wraps the msgpack serialization protocol."""
+    serializer_id = 6  # never change this
+
+    __type_replacements = {}
+
+    def dumpsCall(self, obj, method, vargs, kwargs):
+        return msgpack.packb((obj, method, vargs, kwargs), use_bin_type=True, default=self.default)
+
+    def dumps(self, data):
+        return msgpack.packb(data, use_bin_type=True, default=self.default)
+
+    def loadsCall(self, data):
+        data = self._convertToBytes(data)
+        obj, method, vargs, kwargs = msgpack.unpackb(data, encoding="utf-8", object_hook=self.object_hook)
+        return obj, method, vargs, kwargs
+
+    def loads(self, data):
+        data = self._convertToBytes(data)
+        return msgpack.unpackb(data, encoding="utf-8", object_hook=self.object_hook, ext_hook=self.ext_hook)
+
+    def default(self, obj):
+        replacer = self.__type_replacements.get(type(obj), None)
+        if replacer:
+            obj = replacer(obj)
+        if isinstance(obj, set):
+            return tuple(obj)  # msgpack module can't deal with sets so we make a tuple out of it
+        if isinstance(obj, uuid.UUID):
+            return str(obj)
+        if isinstance(obj, bytearray):
+            return bytes(obj)
+        if isinstance(obj, complex):
+            return msgpack.ExtType(0x30, struct.pack("dd", obj.real, obj.imag))
+        if isinstance(obj, datetime.datetime):
+            if obj.tzinfo:
+                raise errors.SerializeError("msgpack cannot serialize datetime with timezone info")
+            return msgpack.ExtType(0x32, struct.pack("d", obj.timestamp()))
+        if isinstance(obj, datetime.date):
+            return msgpack.ExtType(0x33, struct.pack("l", obj.toordinal()))
+        if isinstance(obj, decimal.Decimal):
+            return str(obj)
+        if isinstance(obj, numbers.Number):
+            return msgpack.ExtType(0x31, str(obj).encode("ascii"))     # long
+        return self.class_to_dict(obj)
+
+    def object_hook(self, obj):
+        if "__class__" in obj:
+            return self.dict_to_class(obj)
+        return obj
+
+    def ext_hook(self, code, data):
+        if code == 0x30:
+            real, imag = struct.unpack("dd", data)
+            return complex(real, imag)
+        if code == 0x31:
+            return int(data)
+        if code == 0x32:
+            return datetime.datetime.fromtimestamp(struct.unpack("d", data)[0])
+        if code == 0x33:
+            return datetime.date.fromordinal(struct.unpack("l", data)[0])
+        raise errors.SerializeError("invalid ext code for msgpack: "+str(code))
+
+    @classmethod
+    def register_type_replacement(cls, object_type, replacement_function):
+        cls.__type_replacements[object_type] = replacement_function
+
+
 """The various serializers that are supported"""
 _serializers = {}
 _serializers_by_id = {}
@@ -547,21 +696,21 @@ def get_serializer(name):
     try:
         return _serializers[name]
     except KeyError:
-        raise Pyro4.errors.SerializeError("serializer '%s' is unknown or not available" % name)
+        raise errors.SerializeError("serializer '%s' is unknown or not available" % name)
 
 
 def get_serializer_by_id(sid):
     try:
         return _serializers_by_id[sid]
     except KeyError:
-        raise Pyro4.errors.SerializeError("no serializer available for id %d" % sid)
+        raise errors.SerializeError("no serializer available for id %d" % sid)
 
 # determine the serializers that are supported
 try:
     import cPickle as pickle
 except ImportError:
     import pickle
-assert Pyro4.config.PICKLE_PROTOCOL_VERSION >= 2, "pickle protocol needs to be 2 or higher"
+assert config.PICKLE_PROTOCOL_VERSION >= 2, "pickle protocol needs to be 2 or higher"
 _ser = PickleSerializer()
 _serializers["pickle"] = _ser
 _serializers_by_id[_ser.serializer_id] = _ser
@@ -569,12 +718,19 @@ import marshal
 _ser = MarshalSerializer()
 _serializers["marshal"] = _ser
 _serializers_by_id[_ser.serializer_id] = _ser
+try:
+    import dill
+    _ser = DillSerializer()
+    _serializers["dill"] = _ser
+    _serializers_by_id[_ser.serializer_id] = _ser
+except ImportError:
+    pass
 try:
     try:
         import importlib
-        json = importlib.import_module(Pyro4.config.JSON_MODULE)
+        json = importlib.import_module(config.JSON_MODULE)
     except ImportError:
-        json = __import__(Pyro4.config.JSON_MODULE)
+        json = __import__(config.JSON_MODULE)
     _ser = JsonSerializer()
     _serializers["json"] = _ser
     _serializers_by_id[_ser.serializer_id] = _ser
@@ -587,14 +743,21 @@ try:
     else:
         ver = serpent.__version__
     ver = tuple(map(int, ver.split(".")))
-    if ver < (1, 11):
-        raise RuntimeError("requires serpent 1.11 or better")
+    if ver < (1, 23):  # serpent 1.23 required
+        raise RuntimeError("requires serpent 1.23 or better")
     _ser = SerpentSerializer()
     _serializers["serpent"] = _ser
     _serializers_by_id[_ser.serializer_id] = _ser
 except ImportError:
     log.warning("serpent serializer is not available")
     pass
+try:
+    import msgpack
+    _ser = MsgpackSerializer()
+    _serializers["msgpack"] = _ser
+    _serializers_by_id[_ser.serializer_id] = _ser
+except ImportError:
+    pass
 del _ser
 
 
@@ -609,7 +772,7 @@ def getAttribute(obj, attr):
         raise AttributeError("attempt to access private attribute '%s'" % attr)
     else:
         obj = getattr(obj, attr)
-    if not Pyro4.config.REQUIRE_EXPOSE or getattr(obj, "_pyroExposed", False):
+    if not config.REQUIRE_EXPOSE or getattr(obj, "_pyroExposed", False):
         return obj
     raise AttributeError("attempt to access unexposed attribute '%s'" % attr)
 
@@ -624,9 +787,8 @@ def fixIronPythonExceptionForPickle(exceptionObject, addAttributes):
     """
     Function to hack around a bug in IronPython where it doesn't pickle
     exception attributes. We piggyback them into the exception's args.
-    Bug report is at http://ironpython.codeplex.com/workitem/30805
-    migrated to github as https://github.com/IronLanguages/main/issues/943
-    Bug is still present in Ironpython 2.7.5
+    Bug report is at https://github.com/IronLanguages/main/issues/943
+    Bug is still present in Ironpython 2.7.7
     """
     if hasattr(exceptionObject, "args"):
         if addAttributes:
@@ -645,7 +807,16 @@ def fixIronPythonExceptionForPickle(exceptionObject, addAttributes):
                     exceptionObject.__dict__.update(piggyback)
 
 
-def get_exposed_members(obj, only_exposed=True):
+__exposed_member_cache = {}
+
+def reset_exposed_members(obj, only_exposed=True, as_lists=False):
+    """Delete any cached exposed members forcing recalculation on next request"""
+    if not inspect.isclass(obj):
+        obj = obj.__class__
+    cache_key = (obj, only_exposed, as_lists)
+    __exposed_member_cache.pop(cache_key, None)
+
+def get_exposed_members(obj, only_exposed=True, as_lists=False, use_cache=True):
     """
     Return public and exposed members of the given object's class.
     You can also provide a class directly.
@@ -655,9 +826,15 @@ def get_exposed_members(obj, only_exposed=True):
     The return value consists of the exposed methods, exposed attributes, and methods
     tagged as @oneway.
     (All this is used as meta data that Pyro sends to the proxy if it asks for it)
+    as_lists is meant for python 2 compatibility.
     """
     if not inspect.isclass(obj):
         obj = obj.__class__
+
+    cache_key = (obj, only_exposed, as_lists)
+    if use_cache and cache_key in __exposed_member_cache:
+        return __exposed_member_cache[cache_key]
+
     methods = set()  # all methods
     oneway = set()  # oneway methods
     attrs = set()  # attributes
@@ -668,7 +845,7 @@ def get_exposed_members(obj, only_exposed=True):
         if inspect.ismethod(v) or inspect.isfunction(v):
             if getattr(v, "_pyroExposed", not only_exposed):
                 methods.add(m)
-                # check if the method is marked with the @Pyro4.oneway decorator:
+                # check if the method is marked with the 'oneway' decorator:
                 if getattr(v, "_pyroOneway", False):
                     oneway.add(m)
         elif inspect.isdatadescriptor(v):
@@ -682,11 +859,17 @@ def get_exposed_members(obj, only_exposed=True):
         # This automatically solves the protection/security issue: you have to
         # explicitly decide to make an attribute into a @property (and to @expose it
         # if REQUIRE_EXPOSED=True) before it is remotely accessible.
-    return {
+    if as_lists:
+        methods = list(methods)
+        oneway = list(oneway)
+        attrs = list(attrs)
+    result = {
         "methods": methods,
         "oneway": oneway,
         "attrs": attrs
     }
+    __exposed_member_cache[cache_key] = result
+    return result
 
 
 def get_exposed_property_value(obj, propname, only_exposed=True):
